// EigenLD.c : Defines the entry point for the console application.

//				Tests the Eigen function

#include <stdlib.h>
#include <stdbool.h>	// for bool
#include "Algo358.h"
#include <time.h>

void Eigen(double _Complex a[][maxA], int n, double _Complex va[], double _Complex ve[][maxA], bool herm, bool *error);
void CSVD(double _Complex a[][maxA], int m, int n, int p, int nu, int nv,
		double s[], double _Complex u[][maxA], double _Complex v[][maxA]);
void TestEigen();
void TimeStamp();
void WriteMatrix(double _Complex [][maxA], int, int);
void WriteMatrixZ(double _Complex [][maxA], int, int);
void MatMulSq(int n, double _Complex a[][maxA], double _Complex b[][maxA], double _Complex p[][maxA]);
void Hermitian(double _Complex a[][maxA], int n);

int ti;			// test case number, 1 to 4
double time0;

// ------- initialize
void init(void)
{
	char name[]    = "EigenLD - Computes Eigenvalues and Eigenvectors of a square complex matrix";
	char version[] = "Lcc-Win32 Console Version 1.1.1";
	char author1[] = "Copyright (c) 1981-2007 by author: Harry J. Smith,";
	char author2[] = "19628 Via Monte Dr., Saratoga, CA 95070.  All rights reserved.";
	char usage1[]  = "Usage: EigenLD [1 to 4] [>Test.out]";
	printf("\n%s\n", name);		// show credits
	printf("%s\n", version);
	printf("%s\n", author1);
	printf("%s\n\n", author2);
	printf("%s\n\n", usage1);	// show usage
} // init

// The main entry point for the application EigenLD
//int _tmain(int argc, _TCHAR* argv[])
int main(int argc, char* argv[])
{
	init();
	char tc[5] = "1";
	if (argc > 1)							// if arguments
		strncpy(tc, argv[1], 2);		// Copy string from user input line
	ti = atoi(tc);
	if (ti < 2 || ti > 4)
		ti = 1;
	printf("Running test case %i of 4\n\n", ti);
	time0 = (double)clock() / (double)CLOCKS_PER_SEC;
	TimeStamp();
	printf("\nEigenLD:\n");
	TestEigen();
	printf("EigenLD:\n");
	printf("Normal end of execution\n\n");
	TimeStamp();
	return 0;
} // main

double _Complex SetTo(double x, double y)		// Set Real and Imag parts
{
	return x + I * y;
} // SetTo

void TestEigen()		//test Eigen
{
	int i, j, n;
	bool herm;
	bool error;
	double _Complex a[maxA][maxA];
	double _Complex u[maxA][maxA];
	double _Complex v[maxA][maxA];
	double _Complex w[maxA][maxA];
	double _Complex va[maxA];
	double _Complex ve[maxA][maxA];
	double _Complex cZero = SetTo(0.0, 0.0);
	a[0][0] = SetTo( 0.4499, -0.1267);
	a[0][1] = SetTo(-0.2361,  0.0775);
	a[0][2] = SetTo( 0.5008, -0.7799);
	a[0][3] = SetTo(-0.7702, -0.3143);
	a[0][4] = SetTo(-0.5084,  0.5621);
	a[0][5] = SetTo(-0.5577, -0.0913);
	a[0][6] = SetTo(-0.7207, -0.9798);
	a[0][7] = SetTo(-0.5367,  0.4677);
	a[1][0] = SetTo(-0.8432, -0.3443);
	a[1][1] = SetTo( 0.0186, -0.6332);
	a[1][2] = SetTo( 0.3505,  0.0166);
	a[1][3] = SetTo(-0.8892,  0.2657);
	a[1][4] = SetTo(-0.5066,  0.6005);
	a[1][5] = SetTo( 0.1421,  0.0154);
	a[1][6] = SetTo(-0.2327, -0.2606);
	a[1][7] = SetTo(-0.4535,  0.4458);
	a[2][0] = SetTo( 0.5896,  0.2601);
	a[2][1] = SetTo( 0.8928,  0.0103);
	a[2][2] = SetTo( 0.4350, -0.2666);
	a[2][3] = SetTo(-0.7799, -0.5512);
	a[2][4] = SetTo(-0.1022, -0.4500);
	a[2][5] = SetTo(-0.6975, -0.8781);
	a[2][6] = SetTo( 0.3364, -0.8415);
	a[2][7] = SetTo( 0.3890,  0.6854);
	a[3][0] = SetTo( 0.3911,  0.3234);
	a[3][1] = SetTo(-0.5605,  0.7638);
	a[3][2] = SetTo(-0.2009,  0.2707);
	a[3][3] = SetTo( 0.0314, -0.4336);
	a[3][4] = SetTo(-0.1046,  0.3267);
	a[3][5] = SetTo(-0.8820, -0.9251);
	a[3][6] = SetTo( 0.4480, -0.0054);
	a[3][7] = SetTo( 0.5823, -0.2339);
	a[4][0] = SetTo(-0.1395, -0.1561);
	a[4][1] = SetTo( 0.3064,  0.0263);
	a[4][2] = SetTo(-0.0975,  0.9019);
	a[4][3] = SetTo( 0.2491,  0.5787);
	a[4][4] = SetTo( 0.4106, -0.8102);
	a[4][5] = SetTo(-0.7131,  0.5675);
	a[4][6] = SetTo( 0.0759,  0.1507);
	a[4][7] = SetTo( 0.1801, -0.3816);
	a[5][0] = SetTo(-0.5889, -0.0238);
	a[5][1] = SetTo( 0.9861, -0.4593);
	a[5][2] = SetTo(-0.5219,  0.8539);
	a[5][3] = SetTo( 0.6382, -0.0232);
	a[5][4] = SetTo( 0.8741,  0.6109);
	a[5][5] = SetTo(-0.2208, -0.1206);
	a[5][6] = SetTo(-0.0868,  0.7716);
	a[5][7] = SetTo(-0.4976,  0.4122);
	a[6][0] = SetTo(-0.3180,  0.8621);
	a[6][1] = SetTo( 0.3905, -0.8872);
	a[6][2] = SetTo( 0.8140, -0.2542);
	a[6][3] = SetTo(-0.2290,  0.4111);
	a[6][4] = SetTo(-0.9723, -0.4355);
	a[6][5] = SetTo( 0.5195,  0.4954);
	a[6][6] = SetTo(-0.8843,  0.4552);
	a[6][7] = SetTo( 0.9156, -0.5878);
	a[7][0] = SetTo( 0.1596, -0.2962);
	a[7][1] = SetTo( 0.0272,  0.3327);
	a[7][2] = SetTo(-0.5806, -0.1952);
	a[7][3] = SetTo(-0.9280, -0.5908);
	a[7][4] = SetTo( 0.6263,  0.0384);
	a[7][5] = SetTo( 0.0210, -0.7325);
	a[7][6] = SetTo(-0.9907, -0.9717);
	a[7][7] = SetTo( 0.0878, -0.9056);
	if (ti == 1)
	{
		n = 5;
		herm = false;
	}
	if (ti == 2)
	{
		n = 5;
		herm = true;
		Hermitian(a, n);	// Set the matrix
	}
	if (ti == 3)
	{
		n = 8;
		herm = false;
	}
	if (ti == 4)
	{
		n = 8;
		Hermitian(a, n);	// Set the matrix
		herm = true;
	}
		printf("Call the Eigen function to compute the Eigenvalues and Eigenvectors of\n");
	if (herm)
		printf("an N by N Hermitian");
	else
		printf("an N by N complex");
	printf(" matrix A, N = %i.\n", n);
	printf("\nMatrix A:\n\n");
	WriteMatrix(a, n, n);
	int p = 0;
	int nu = n;
	int nv = n;

	// INITIAL TESTS
	void PolNorm(double _Complex a[], int *n);
	double _Complex pp[9];
	pp[0] = SetTo( 0.0000,  0.0000);
	pp[1] = SetTo(-0.2009,  0.2707);
	pp[2] = SetTo( 0.4106, -0.8102);
	int ppn = 3;
	PolNorm(pp, &ppn);
	void Laguer(double _Complex a[], int n, int *its, double _Complex *x);
	pp[0] = SetTo( 0.4499, -0.1267);
	pp[1] = SetTo(-0.2009,  0.2707);
	pp[2] = SetTo( 0.4106, -0.8102);
	int its;
	// answer 0.80826026, 1.0269444
	double _Complex root;
	bool polish = true;
	Laguer(pp, 3, &its, &root);
	void PolRoots(double _Complex p[], int n, bool polish, double _Complex r[]);
	double _Complex roots[2];
	PolRoots(pp, 3, polish, roots);
	void CharPoly(double _Complex a[][maxA], int n, double _Complex p[]);
	CharPoly(a, n, pp);
	// END OF INITIAL TESTS

	Eigen(a, n, va, ve, herm, &error);
	printf("Eigenvalues va:\n\n");
	for (i = 0; i < n; i++)
		printf("  (%.14f,%.14f)\n", creal(va[i]), cimag(va[i]));
	printf("\nEigenvectors ve:\n\n");
	WriteMatrix(ve, n, n);
	MatMulSq(n, a, ve, u);
	printf("A * ve:\n\n");
	WriteMatrix(u, n, n);
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			w[i][j] = cZero;
	for (i = 0; i < n; i++)
	{
		double _Complex tw = va[i];		// Lcc-Win32 had a problem with w[i][i] = va[i];
		w[i][i] = tw;
	}
	MatMulSq(n, ve, w, v);
	double _Complex maxErr = cZero;
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
		{
			w[i][j] = u[i][j] - v[i][j];
			if (fabs(creal(w[i][j])) > creal(maxErr))
				maxErr = SetTo(fabs(creal(w[i][j])), cimag(maxErr));
			if (fabs(cimag(w[i][j])) > cimag(maxErr))
				maxErr = SetTo(creal(maxErr), fabs(cimag(w[i][j])));
		}
	printf("A * ve - ve * Diag(va), should be zero: Max error = (%.3e,%.3e)\n\n", creal(maxErr), cimag(maxErr));
	WriteMatrixZ(w, n, n);
} // TestEigen

void Hermitian(double _Complex a[][maxA], int n)
// make a an n by n Hermitian matrix. use upper triangle to make lower triangle
{
	int i, j, k;
	for (i = 0; i < n - 1; i++)
	{
		for (j = i + 1; j < n; j++)
			a[j][i] = conj(a[i][j]);
	}
	for (k = 0; k < n; k++)
		a[k][k] = SetTo(creal(a[k][k]), 0.0);

} // Hermitian

void MatMulSq(int n, double _Complex a[][maxA], double _Complex b[][maxA], double _Complex p[][maxA])
// p = a * b. a, b, p all n by n complex matrices
{
	int i, j, k;
	double _Complex cZero = SetTo(0.0, 0.0);
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
		{
			p[i][j] = cZero;
			for (k = 0; k < n; k++)
			{
				p[i][j] = p[i][j] + a[i][k] * b[k][j];
			}
		}
} // MatMulSq

void WriteMatrix(double _Complex a[][maxA], int m, int n)
// print an m by n complex matrix
{
	int i, j;
	for (i = 0; i < m; i++)
	{
		for (j = 0; j < n; j++)
		{
			printf(" ");
			if (creal(a[i][j]) >= 0.0)
				printf(" ");
			else
				printf("-");
			if (creal(a[i][j]) == 0.0)
				printf("0.0000");
			else
				printf("%.4f", fabs(creal(a[i][j])));
			if (cimag(a[i][j]) >= 0.0)
				printf("+");
			else
				printf("-");
			if (cimag(a[i][j]) == 0.0)
				printf("0.0000");
			else
				printf("%.4f", fabs(cimag(a[i][j])));
		}
		printf("\n");
	}
	printf("\n");
} // WriteMatrix

void WriteMatrixZ(double _Complex a[][maxA], int m, int n)
// print an m by n complex matrix that is near zero
{
	int i, j;
	for (i = 0; i < m; i++)
	{
		for (j = 0; j < n; j++)
		{
			printf(" ");
			if (creal(a[i][j]) >= 0.0)
				printf(" ");
			else
				printf("-");
			if (creal(a[i][j]) == 0.0)
				printf("0.0000");
			else
				printf("%.0e", fabs(creal(a[i][j])));
			if (cimag(a[i][j]) >= 0.0)
				printf("+");
			else
				printf("-");
			if (cimag(a[i][j]) == 0.0)
				printf("0.0000");
			else
				printf("%.0e", fabs(cimag(a[i][j])));
		}
		printf("\n");
	}
	printf("\n");
} // WriteMatrixZ

void TimeStamp()
// printout the current MM/dd/yyyy hh:mm:ss AM/PM  s.ss sec. as a timestamp
{
	int i;
	double time1 = (double)clock() / (double)CLOCKS_PER_SEC;
	double timeD = time1 - time0;
	time0 = time1;
	char date[9];
	struct tm *newtime;
	time_t ltime;
	/* get time in seconds */
	time(&ltime);
	/* convert it to the structure tm */
	newtime = localtime(&ltime);
	char sta[30];
	for (i = 0; i < 24; i++)
	{
		sta[i] = (char)asctime(newtime)[i];
	}
	sta[24] = 0;
	printf("%s  %.2f sec.\n", sta, timeD);
} // TimeStamp

/* end of file EigenLD.c ------------------------------------------ */
